import copy
import json
import time
import uuid
import hashlib
import pyrql
from jsonschema import validate
from pyrql.query import And, Or, Filter

from s3i import EventMessage


class NamedEvent(object):
    def __init__(self, topic, json_schema):
        self.__topic = topic
        self.__json_schema = json_schema
        self.__msg = {}

    @property
    def topic(self):
        return self.__topic

    @property
    def json_schema(self):
        return self.__json_schema

    @property
    def msg(self):
        return self.__msg

    def generate_event_msg(self, content):
        validate(instance=content, schema=self.__json_schema)
        _msg = EventMessage()
        _msg.fillEventMessage(
            sender=self.__topic.split(".")[0],
            message_id="s3i:{}".format(uuid.uuid4()),
            topic=self.topic,
            timestamp=int(time.time() * 1000.0),
            content=content
        )
        self.__msg = _msg.base_msg


class CustomEvent(object):
    def __init__(self, filter_object, filter_attribute_paths,
                 attribute_paths, topic, json_entry):
        self.__filter_object = filter_object
        self.__filter_attribute_paths = filter_attribute_paths
        self.__attribute_paths = attribute_paths
        self.__topic = topic
        self.__is_triggered = False
        self.__msg = {}
        self.__json_entry = json_entry
        self.__cur_attributes_dict = {}
        self.__content_old = {}

    @property
    def filter_object(self):
        return self.__filter_object

    @property
    def filter_attribute_paths(self):
        return self.__filter_attribute_paths

    @property
    def attribute_paths(self):
        return self.__attribute_paths

    @property
    def topic(self):
        return self.__topic

    @property
    def msg(self):
        return self.__msg

    def check_filter(self):
        """
        Check the rql filter being triggered
        """
        if self.check_filter_attributes_changed():
            return self.__filter_object(self.__cur_attributes_dict)
        return False 

    def check_filter_attributes_changed(self):
        new_cur_dict = {}
        for path in self.filter_attribute_paths:
            new_cur_dict[path.replace("/", " ")] = _uriToData(path, self.__json_entry)
            if new_cur_dict != self.__cur_attributes_dict:
                self.__cur_attributes_dict = copy.deepcopy(new_cur_dict)
                return True
        return False

    def generate_event_msg(self):
        _content = {}
        for i in self.__attribute_paths:
            var = _uriToData(i, self.__json_entry)
            _content[i] = var
        if _content == self.__content_old:
            return
        _msg = EventMessage()
        _msg.fillEventMessage(
            sender=self.__topic.split(".")[0],
            message_id="s3i:{}".format(uuid.uuid4()),
            topic=self.__topic,
            timestamp=int(time.time() * 1000.0),
            content=_content
        )
        self.__msg = _msg.base_msg
        self.__content_old = _content


class EventManager:
    def __init__(self, json_entry):
        self.__json_entry = json_entry
        self.__named_event_dict = {}
        self.__custom_event_dict = {}

    @property
    def json_entry(self):
        return self.__json_entry

    @property
    def named_event_dict(self):
        return self.__named_event_dict

    @property
    def custom_event_dict(self):
        return self.__custom_event_dict

    @json_entry.setter
    def json_entry(self, value):
        if isinstance(value, dict):
            self.__json_entry = value

    def add_named_event(self, topic, json_schema):
        """
        Add a namedEvent to EventManager

        """
        if topic in self.__named_event_dict.keys():
            return False
        _named_event = NamedEvent(topic, json_schema)
        self.__named_event_dict[topic] = _named_event
        return True

    def delete_named_event(self, topic):
        if topic not in self.__named_event_dict.keys():
            return False
        _deleted_obj = self.__named_event_dict.pop(topic)
        del _deleted_obj

    def add_custom_event(self, rql_expression, attribute_paths):
        rql = RQL(rql_expression=rql_expression)
        if not rql.is_rql_valid:
            return False, None

        for path in rql.get_attributes():
            if _uriToData(path, self.__json_entry) is None:
                return False, None

        for path in attribute_paths:
            if _uriToData(path, self.__json_entry) is None:
                return False, None

        rql_filter = rql.make_filter()
        topic = "{}.{}".format(self.__json_entry.get("thingId"),
                               hashlib.md5((rql_expression + str(attribute_paths)).encode("utf-8")).hexdigest())
        _custom_event = CustomEvent(filter_object=rql_filter, filter_attribute_paths=rql.get_attributes(),
                                    attribute_paths=attribute_paths, topic=topic, json_entry=self.__json_entry)
        self.__custom_event_dict[topic] = _custom_event
        return True, topic

    def delete_custom_event(self, topic):
        if topic not in self.__custom_event_dict.keys():
            return False
        _deleted_obj = self.__custom_event_dict.pop(topic)
        del _deleted_obj
        return True

    def emit_named_event(self, publisher, topic, content):

        event_obj = self.__named_event_dict.get(topic)

        if isinstance(event_obj, NamedEvent):
            event_obj.generate_event_msg(content=content)
            publisher.publish_event(
                topic=topic,
                msg=json.dumps(event_obj.msg)
            )

    def emit_custom_event(self, publisher, topic):
        event_obj = self.__custom_event_dict.get(topic)
        if isinstance(event_obj, CustomEvent):
            if event_obj.check_filter():
                event_obj.generate_event_msg()
                publisher.publish_event(
                    msg=json.dumps(event_obj.msg),
                    topic=topic
                )


class RQL:
    """
    class for parsing, creating and checking rql filter
    """

    def __init__(self, rql_expression):
        """
        Constructor
        """

        self.__rql_expression = rql_expression
        self.__parsed_rql_expression = None
        self.__is_rql_valid = False

        self.__rql_logic_operator = ["and", "or"]
        self.__rql_math_operator = ["eq", "ne", "gt", "ge", "lt", "ge", "like", "exists"]

        self.__parse()

    @property
    def rql_expression(self):
        return self.__rql_expression

    @property
    def parsed_rql_expression(self):
        return self.__parsed_rql_expression

    @property
    def is_rql_valid(self):
        return self.__is_rql_valid

    def __parse(self):
        """
        Function to parse rql expression. If the expression is invalid, return empty dictionary

        :param rql_expr: rql expression
        :type rql_expr: str

        :return: parsed rql expression
        :rtype: dict

        """
        try:
            self.__parsed_rql_expression = pyrql.parse(self.__rql_expression.replace("/", " "))
            self.__is_rql_valid = True
        except pyrql.RQLSyntaxError:
            self.__parsed_rql_expression = {}
            self.__is_rql_valid = False

    def make_filter(self):
        """
        Function to create a RQL filter object. If the RQL expression is invalid, return None.

        :return: RQL filter

        Returns:

        """
        if self.__parsed_rql_expression.get("name") in self.__rql_math_operator:
            rql_filter = Filter(self.__parsed_rql_expression.get("name"),
                                self.__parsed_rql_expression.get("args")[0],
                                self.__parsed_rql_expression.get("args")[1])

        elif self.__parsed_rql_expression.get("name") in self.__rql_logic_operator:
            args = self.__parsed_rql_expression.get("args")
            rql_filter = self.__create_single_filter(args)
            for i in range(len(rql_filter) - 1):
                if self.__parsed_rql_expression.get("name") == "and":
                    rql_filter = And(rql_filter[i], rql_filter[i + 1])
                if self.__parsed_rql_expression.get("name") == "or":
                    rql_filter = Or(rql_filter[i], rql_filter[i + 1])
        else:
            return None
        return rql_filter

    def __create_single_filter(self, args, filters=None):
        """
        Function to create a single filter which only contains one related variable in filter expression

        """
        if filters is None:
            filters = []
        for i in range(len(args)):
            if args[i].get("name") not in self.__rql_logic_operator:
                filters.append(
                    Filter(args[i].get("name"), args[i].get("args")[0], args[i].get("args")[1]))
            else:
                filters.append([])
                self.__create_single_filter(args=args[i].get("args"), filters=filters[i])

                if args[i].get("name") == "and":

                    for j in range(len(filters[i]) - 1):
                        filters[i] = And(filters[i][j], filters[i][j + 1])

                if args[i].get("name") == "or":

                    for j in range(len(filters[i]) - 1):
                        filters[i] = Or(filters[i][j], filters[i][j + 1])

        return filters

    def get_attributes(self):
        """
        Function to get all attributes contained in the rql expression.
        If the rql expression is invalid, return empty list

        :return: attributes
        :rtype: list

        """
        _attributes = list()
        if self.__parsed_rql_expression.get("name") in self.__rql_math_operator:
            _attributes.append(self.__parsed_rql_expression.get("args")[0])

        elif self.__parsed_rql_expression.get("name") in self.__rql_logic_operator:
            args = self.__parsed_rql_expression.get("args")
            _attributes = self._get_attributes(args)

        for i in range(len(_attributes)):
            _attributes[i] = _attributes[i].replace(" ", "/")

        return _attributes

    def _get_attributes(self, args):
        _attributes = []
        for i in range(len(args)):
            if args[i].get("name") not in self.__rql_logic_operator:
                _attributes.append(args[i].get("args")[0])
            else:
                _temp_attributes = self._get_attributes(args=args[i].get("args"))
                for temp in _temp_attributes:
                    _attributes.append(temp)
        return _attributes


resGetValue = []


def _uriToData(uri, ml40_model):
    """Returns a copy of the value found at uri.

    :param uri: Path to value
    :rtype: Feature

    """
    global resGetValue
    if uri == "":
        return ml40_model
    else:
        uri_list = uri.split("/")
        if uri_list[0] == "features":
            try:
                return ml40_model[uri]
            except KeyError:
                return None

        try:
            _getValue(ml40_model, uri_list, ml40_model)
        except Exception:
            return None
        if resGetValue.__len__() == 0:
            return None
        response = copy.deepcopy(resGetValue)
        resGetValue.clear()
        if response.__len__() == 1:
            return response[0]
        else:
            return response


def _getValue(source, uri_list, ml40_model):
    """Searches for the value specified by uri_list in source and stores
    the result in __resGetValue.

    :param source: Object that is scanned
    :param uri_list: List containing path

    """
    global resGetValue
    value = source[uri_list[0]]
    if uri_list.__len__() == 1:
        # if is ditto-feature
        if isinstance(value, str):
            try:
                stringValue_split = value.split(":")
                if stringValue_split[0] == "ditto-feature":
                    value = ml40_model["features"][stringValue_split[1]][
                        "properties"
                    ][uri_list[0]]
            except:
                pass
        resGetValue.append(value)
        return
    if isinstance(value, dict):
        # ??? uri_list.pop(0) better?!
        del uri_list[0]
        _getValue(value, uri_list, ml40_model)
    if isinstance(value, list):
        if isinstance(value[0], (str, int, float, bool, list)):
            return value
        if isinstance(value[0], dict):
            for item in value:
                if item["class"] == "ml40::Thing":
                    for i in item["roles"]:
                        if _findValue(i, uri_list[1]):
                            uri_list_1 = copy.deepcopy(uri_list)
                            del uri_list_1[:2]
                            _getValue(item, uri_list_1, ml40_model)
                    _f = _findValue({"identifier": item.get("identifier")}, uri_list[1]) or \
                         _findValue({"name": item.get("name")}, uri_list[1])
                    if _f:
                        uri_list_1 = copy.deepcopy(uri_list)
                        del uri_list_1[:2]
                        _getValue(item, uri_list_1, ml40_model)
                else:
                    if _findValue(item, uri_list[1]):
                        uri_list_1 = copy.deepcopy(uri_list)
                        del uri_list_1[:2]
                        if not uri_list_1:
                            resGetValue.append(item)
                            return
                        else:
                            _getValue(item, uri_list_1, ml40_model)
    if isinstance(value, (str, int, float, bool)):
        # if is ditto-feature
        if isinstance(value, str):
            try:
                stringValue_split = value.split(":")
                if stringValue_split[0] == "ditto-feature":
                    value = ml40_model["features"][stringValue_split[1]][
                        "properties"
                    ][uri_list[0]]
            except:
                pass
        resGetValue.append(value)


def _findValue(json, value):
    """Returns true if value has been found in json, otherwise returns false.

    :param json: dictionary
    :param value:
    :returns:
    :rtype:

    """

    # TODO: Simplify: value in json.values()
    for item in json:
        if json[item] == value:
            # print("Parameter: ", json[item])
            return True
    return False



import requests
import json
import os
import pgpy
from s3i.ditto_manager import DittoManager
from s3i.exception import S3IDirectoryError, raise_error_from_response


class Directory(DittoManager):
    """Class Directory contains functions to query things in S3I Directory """

    def __init__(self, s3i_dir_url, token):
        """
        Constructor

        :param s3i_dir_url: url of S3I Directory, https://dir.s3i.vswf.dev/api/2/
        :type s3i_dir_url: str
        :param token: id token obtained from S3I IdP
        :type token: str
        """
        super().__init__(s3i_dir_url, token)
    
    def queryEndpointService(self, thingID, serviceType):
        """
        Query a endpoint of service

        :param thingID: thingId
        :type thingID: str
        :param serviceType: service type
        :type serviceType: str
        """
        thing = self.queryThingIDBased(thingID=thingID)
        thing = raise_error_from_response(thing, S3IDirectoryError, 200)
        thing_links = thing["attributes"]["thingStructure"]["links"]
        endpoints = list()
        for link in thing_links:
            [endpoints.append(service["endpoints"]) for service in link["target"]
             ["services"] if service["serviceType"] == serviceType]
        return endpoints


    def getPublicKey(self, thingID):  # TODO thingIDs is list
        """
        Query public key of a thing

        :param thingID: thingId
        :type thingID: str
        """
        if isinstance(thingID, str):
            response = requests.get(url=self.ditto_url+"things/"+thingID +
                                    "/attributes/publicKey", headers=self.ditto_headers)
            raise_error_from_response(response, S3IDirectoryError, 200)
            return response.text
        if isinstance(thingID, list):
            keys = list()
            for id in thingID:
                response = requests.get(url=self.ditto_url+"things/"+id +
                                        "/attributes/publicKey", headers=self.ditto_headers)
                raise_error_from_response(response, S3IDirectoryError, 200)
                keys.append(response.text)
            return keys
        else:
            print(
                "[S3I]: Try to get public keys from thing ids but thing ids are neither strings nor lists.")
            return

    def putPublicKey(self, thingID, file):
        """
        Upload/Update public key of a thing

        :param thingID: thingId
        :type thingID: str
        :param file: file which stores a public key
        :type file: str
        """
        path = os.path.join(os.getcwd(), file)
        key, _ = pgpy.PGPKey.from_file(path)
        keystr = key.__str__()
        url_str = self.ditto_url+"things/" + thingID + "/attributes/publicKey"
        response = requests.put(
            url=url_str,  headers=self.ditto_headers, data=json.dumps(keystr))
        raise_error_from_response(response, S3IDirectoryError, 204)
        return response



import requests
import json
from schema import Schema, And, Or, Optional, SchemaError
import pika.exceptions as s3ib


class S3IError(Exception):
    def __init__(self, error_msg):
        Exception.__init__(self, error_msg)
        self.error_msg = error_msg

    def __str__(self):
        return str(self.error_msg)


class S3IDittoError(S3IError):
    def __init__(self, error_msg):
        S3IError.__init__(self, error_msg)
        self.error_msg = error_msg

    def __str__(self):
        if isinstance(self.error_msg, dict):
            return str(self.error_msg.get("message"))
        else:
            return str(self.error_msg)


class S3IDirectoryError(S3IDittoError):
    def __init__(self, error_msg):
        S3IDittoError.__init__(self, error_msg)
        self.error_msg = error_msg

    def __str__(self):
        def __str__(self):
            if isinstance(self.error_msg, dict):
                return str(self.error_msg.get("message"))
            else:
                return str(self.error_msg)


class S3IRepositoryError(S3IError):
    pass


class S3IIdentityProviderError(S3IError):
    def __init__(self, error_msg):
        Exception.__init__(self, error_msg)
        self.error_msg = error_msg

    def __str__(self):
        if isinstance(self.error_msg, dict):
            return str(self.error_msg.get("error"))
        else:
            return str(self.error_msg)


class S3IBMessageError(S3IError):
    def __init__(self, error_msg):
        S3IError.__init__(self, error_msg)
        self.error_msg = error_msg

    def __str__(self):
        return str(self.error_msg)


class S3IBrokerRESTError(S3IError):
    def __init__(self, error_msg):
        S3IError.__init__(self, error_msg)
        self.error_msg = error_msg

    def __str__(self):
        if isinstance(self.error_msg, dict):
            return str(self.error_msg.get("error"))
        elif isinstance(self.error_msg, list):
            for i in self.error_msg:
                return str(i.get("error"))


class S3IBrokerAMQPError(S3IError):
    def __init__(self, error_msg):
        S3IError.__init__(self, error_msg)
        self.error_msg = error_msg

    def __str__(self):
        if self.error_msg is None:
            return "unknown error"
        else:
            return str(self.error_msg)


def raise_error_from_s3ib_amqp(function, error, *args, **kwargs):
    try:
        r = function(*args, **kwargs)

    except (s3ib.AMQPChannelError, s3ib.AMQPConnectionError, s3ib.AMQPError,
            s3ib.AMQPHeartbeatTimeout, s3ib.AuthenticationError, s3ib.BodyTooLongError,
            s3ib.ChannelClosed, s3ib.ChannelClosedByBroker, s3ib.ChannelClosedByClient,
            s3ib.ChannelError, s3ib.ChannelWrongStateError, s3ib.ConnectionBlockedTimeout,
            s3ib.ConnectionClosed, s3ib.ConnectionClosedByBroker, s3ib.ConnectionClosedByClient,
            s3ib.ConnectionOpenAborted, s3ib.ConnectionWrongStateError, s3ib.ConsumerCancelled,
            s3ib.DuplicateConsumerTag, s3ib.DuplicateGetOkCallback, s3ib.IncompatibleProtocolError,
            s3ib.InvalidChannelNumber, s3ib.InvalidFieldTypeException, s3ib.InvalidFrameError,
            s3ib.MethodNotImplemented, s3ib.NackError, s3ib.NoFreeChannels, s3ib.ProbableAccessDeniedError,
            s3ib.ProbableAuthenticationError, s3ib.ProbableAccessDeniedError,
            s3ib.StreamLostError, s3ib.ShortStringTooLong, s3ib.ReentrancyError,
            s3ib.UnexpectedFrameError, s3ib.UnroutableError, s3ib.UnsupportedAMQPFieldException) as e:
        raise error(error_msg=e)
    if r is not None:
        return r


def raise_error_from_s3ib_msg(msg, error):
    user_msg_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i:" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers":  And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "userMessage"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            Optional("attachments"): And([{Optional(str): Or(str, bool, int, float, dict)}], lambda s: len(s) >= 0),
            "subject": And(str),
            "text": And(str)
        }, ignore_extra_keys=False
    )
    service_req_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "serviceRequest"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            "serviceType": str,
            Optional("parameters"): And({Optional(str): Or(str, dict, int, float, bool, list)})
        }, ignore_extra_keys=False
    )
    service_reply_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "serviceReply"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            "serviceType": str,
            Optional("results"): Or({Optional(str): Or(str, dict, int, float, bool, list)}, str, list, bool, float, int)
        }, ignore_extra_keys=False
    )
    get_value_request_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "getValueRequest"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            Optional("timeInterval"): {"start": int, "end": int},
            "attributePath": str
        }, ignore_extra_keys=False
    )
    get_value_reply_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "getValueReply"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            "value": Or(str, dict, int, float, bool, list)
        }, ignore_extra_keys=False
    )
    set_value_request_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "setValueRequest"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            "attributePath": str,
            "newValue": Or(str, dict, int, float, bool, list)
        }, ignore_extra_keys=False
    )
    set_value_reply_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "setValueReply"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            "ok": bool
        }, ignore_extra_keys=False
    )
    create_attribute_request_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "createAttributeRequest"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            "attributePath": str,
            "newValue": Or(str, int, float, dict, bool, list)
        }, ignore_extra_keys=False
    )
    create_attribute_reply_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "createAttributeReply"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            "ok": bool
        }, ignore_extra_keys=False
    )
    delete_attribute_request_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "deleteAttributeRequest"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            Optional("timeInterval"): {"start": int, "end": int},
            "attributePath": str,
        }, ignore_extra_keys=False
    )
    delete_attribute_reply_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "deleteAttributeReply"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            "ok": bool
        }, ignore_extra_keys=False
    )

    subscribe_custom_event_request_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "subscribeCustomEventRequest"),
            "filter": str,
            "attributePaths": And([str], lambda s: len(s) > 0),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n)
        }, ignore_extra_keys=False
    )

    subscribe_custom_event_reply_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "subscribeCustomEventReply"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            "status": Or(bool, str),
            "topic": str
        }, ignore_extra_keys=False
    )

    unsubscribe_custom_event_request_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "unsubscribeCustomEventRequest"),
            "topic": str,
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n)
        }, ignore_extra_keys=False
    )

    unsubscribe_custom_event_reply_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "receivers": And([And(str, lambda n: "s3i" in n)], lambda s: len(s) > 0),
            "messageType": And(str, lambda s: s == "unsubscribeCustomEventReply"),
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            "status": Or(bool, str),
            "topic": str
        }, ignore_extra_keys=False
    )

    event_message_schema = Schema(
        {
            "sender": And(str, lambda n: "s3i" in n),
            "identifier": And(str, lambda n: "s3i:" in n),
            "messageType": And(str, lambda s: s == "eventMessage"),
            "topic": str,
            "timestamp": int,
            Optional("replyToEndpoint"): And(str, lambda n: "s3ib" in n),
            Optional("replyingToMessage"): And(str, lambda n: "s3i" in n),
            Optional("content"): And({Optional(str): Or(str, dict, int, float, bool, list)})
        }, ignore_extra_keys=False
    )

    try:
        if not isinstance(msg, dict):
            raise error(error_msg="S3IB Message should be formalized as dict")

        msg_type = msg.get("messageType")
        if msg_type == "userMessage":
            return user_msg_schema.validate(msg)
        elif msg_type == "serviceRequest":
            return service_req_schema.validate(msg)
        elif msg_type == "serviceReply":
            return service_reply_schema.validate(msg)
        elif msg_type == "getValueRequest":
            return get_value_request_schema.validate(msg)
        elif msg_type == "getValueReply":
            return get_value_reply_schema.validate(msg)
        elif msg_type == "setValueRequest":
            return set_value_request_schema.validate(msg)
        elif msg_type == "setValueReply":
            return set_value_reply_schema.validate(msg)
        elif msg_type == "createAttributeRequest":
            return create_attribute_request_schema.validate(msg)
        elif msg_type == "createAttributeReply":
            return create_attribute_reply_schema.validate(msg)
        elif msg_type == "deleteAttributeRequest":
            return delete_attribute_request_schema.validate(msg)
        elif msg_type == "deleteAttributeReply":
            return delete_attribute_reply_schema.validate(msg)
        elif msg_type == "subscribeCustomEventRequest":
            return subscribe_custom_event_request_schema.validate(msg)
        elif msg_type == "subscribeCustomEventReply":
            return subscribe_custom_event_reply_schema.validate(msg)
        elif msg_type == "unsubscribeCustomEventRequest":
            return unsubscribe_custom_event_request_schema.validate(msg)
        elif msg_type == "unsubscribeCustomEventReply":
            return unsubscribe_custom_event_reply_schema.validate(msg)
        elif msg_type == "eventMessage":
            return event_message_schema.validate(msg)
        else:
            error_msg = "Invalid type of S3I-B messages"
            raise error(error_msg=error_msg)

    except SchemaError as e:
        raise error(error_msg=e.autos)


def raise_error_from_response(response, error, expected_response_code):
    if response.status_code == expected_response_code:
        if expected_response_code == requests.codes.no_content:
            return {}
        else:
            try:
                return response.json()
            except json.decoder.JSONDecodeError:
                return {}
    else:
        try:
            error_msg = response.json()
        except json.decoder.JSONDecodeError:
            error_msg = response.text
    raise error(error_msg=error_msg)



import pgpy
import os
from datetime import timedelta
from pgpy.constants import PubKeyAlgorithm, KeyFlags, HashAlgorithm, SymmetricKeyAlgorithm, CompressionAlgorithm


class Key:
    """PGP Key as described in RFC 4880"""

    def __init__(self, path_demo_dir=None, filename=None, key_str=None, size=512):
        """This function intializes a PGP key. There are several possibilities to do it: You can specify a path or a key in string format to load an existing key. If neither a path nor a key string is given, a new key is generated.

        :param path_demo_dir: Path to the ASC file (default: None)
        :type path_demo_dir: str
        :param filename: name os the ASC file storing the PGP key
        :type filename: str
        :param key_str: PGP key in string format (default: None)
        :type key_str: str
        :param size: size of the key if an new key is generated (default: 512)
        :type size: int
        """
        if isinstance(path_demo_dir, str):
            self.loadFromFile(path_demo_dir, filename)
        elif isinstance(key_str, str):
            key_list = pgpy.PGPKey.from_blob(
                key_str.replace("\\n", "\n").strip('"'))
            self.key = key_list[0]
        else:
            self.key = pgpy.PGPKey.new(PubKeyAlgorithm.RSAEncryptOrSign, size)

    def generateKey(self, userID, comment="", email=""):
        """This function generates a new PGP key.

        :param userID: User ID name, or photo. If this is a bytearray, it will be loaded as a photo. Otherwise, it will be used as the name field for a User ID.
        :type pn: bytearray, str, unicode  
        :param comment: The comment field for a User ID. Ignored if this is a photo.
        :type comment: str, unicode
        :param email: The email address field for a User ID. Ignored if this is a photo.
        :type email: str, unicode
        :return: PGP key 
        :rtype: pgpy.PGPKey
        """
        self.uid = pgpy.PGPUID.new(userID, comment, email)
        self.key.add_uid(self.uid, usage={KeyFlags.Sign, KeyFlags.EncryptCommunications, KeyFlags.EncryptStorage},
                         hashes=[HashAlgorithm.SHA256, HashAlgorithm.SHA384,
                                 HashAlgorithm.SHA512, HashAlgorithm.SHA224],
                         ciphers=[SymmetricKeyAlgorithm.AES256,
                                  SymmetricKeyAlgorithm.AES192, SymmetricKeyAlgorithm.AES128],
                         compression=[CompressionAlgorithm.ZLIB, CompressionAlgorithm.BZ2, CompressionAlgorithm.ZIP, CompressionAlgorithm.Uncompressed])
        return self.key

    def addKeyExpiration(self, expDays):
        """This function specifies the amount of days the key should be valid.

        :param expDays: Amount of days after which the key should expire.
        :type expDays: int
        """
        self.key.add_uid(self.uid, key_expires=timedelta(days=expDays))

    def exportsecKeyToFile(self, filename="secKey.asc", path = None):
        """This function exports the key to an ASC file in the key folder with the given filename.

        :param filename: filename of the output file
        :type filename: str
        """
        if not path:
            save_path = 'key'
            fp = os.path.dirname(os.path.abspath(__file__))
            path = os.path.join(fp, "..", save_path, filename)
        else:
            path = os.path.join(path,filename)
        f = open(path, "w")
        keystr = str(self.key)
        f.write(keystr)
        f.close()

    def exportPubKeyToFile(self, filename="pubKey.asc", path = None):
        """This function exports the key to an ASC file in the key folder with the given filename.

        :param filename: filename of the output file
        :type filename: str
        """
        if not path:
            save_path = 'key'
            fp = os.path.dirname(os.path.abspath(__file__))
            path = os.path.join(fp, "..", save_path, filename)
        else:
            path = os.path.join(path,filename)
        f = open(path, "w")
        keystr = str(self.key.pubkey)
        f.write(keystr)
        f.close()

    def loadFromFile(self, path_demo_dir, filename):
        """This function loads a PGP key that is given as an ASC file from a subdirectory.

        :param path_demo_dir: Path to the ASC file (default: None)
        :type path_demo_dir: str
        :param filename: name os the ASC file storing the PGP key
        :type filename: str
        """

        path = os.path.join(path_demo_dir, filename)
        self.key, _ = pgpy.PGPKey.from_file(path)



import asyncio
from typing import Dict, Tuple, List, Any, Awaitable, Callable

from websocket import send

from s3i.logger import APP_LOGGER
MsgCallbackType = Callable[[Dict[str, Any]], Awaitable[None]]

class CallbackManager(object):
    """
    Global callback system, which is aimed at to be a single place to manage callbacks and process them
    """
    CALLBACK = 'callback'
    ONE_SHOT = 'one_shot'
    IS_ASYNC_CALLBACK = 'is_async_callback'
    ARGS = 'args'
    KWARGS = 'kwargs'

    def __init__(self):
        """
        Create an instance of the CallbackManager
        """
        self._stack = {}
        self.__message_callbacks: Dict[Tuple[str, str, str, str], List[Tuple[MsgCallbackType, bool]]] = {("", "", "", ""): []}
        self.logger = APP_LOGGER

    def add(self, prefix, callback, one_shot, is_async, *args, **kwargs):
        # prepare the stack
        if prefix not in self._stack:
            self._stack[prefix] = []

        if not isinstance(one_shot, bool):
            raise TypeError
        if not isinstance(is_async, bool):
            raise TypeError

        # create callback dictionary
        callback_dict = self.create_callback_dict(
            callback,
            one_shot,
            is_async,
            *args,
            **kwargs
        )

        # check for a duplicate
        if callback_dict in self._stack[prefix]:
            return prefix

        # append
        self._stack[prefix].append(callback_dict)
        return prefix

    def remove(self, prefix, callback_dict):
        if prefix not in self._stack:
            return False
        for callback in self._stack[prefix]:
            if callback == callback_dict:
                self._stack[prefix].remove(callback)
                return True
        return False

    def clear(self):
        if self._stack:
            self._stack = {}

    def process(self, prefix, loop=asyncio.get_event_loop()):
        if prefix not in self._stack:
            return False

        for callback_dict in self._stack[prefix]:
            method = callback_dict[self.CALLBACK]
            args = callback_dict[self.ARGS]
            kwargs = callback_dict[self.KWARGS]
            if callback_dict[self.IS_ASYNC_CALLBACK]:
                loop.create_task(method(*args, **kwargs))
            else:
                method(*args, **kwargs)

        remove = [callback_dict for callback_dict in self._stack[prefix] if callback_dict[self.ONE_SHOT]]
        for callback_dict in remove:
            self.remove(prefix, callback_dict)

        return True

    def create_callback_dict(self, callback, one_shot, is_async, *args, **kwargs):
        """
        Create and return callback dictionary

        :param method callback:
        :param bool one_shot:
        :param is_async:
        :return:
        """

        return {
            self.CALLBACK: callback,
            self.ONE_SHOT: one_shot,
            self.IS_ASYNC_CALLBACK: is_async,
            self.ARGS: args,
            self.KWARGS: kwargs
        }

    def add_message_callback(
            self,
            callback: MsgCallbackType,
            sender_id: str = "",
            receiver_id: str = "",
            message_type: str = "",
            replying_to_msg: str = "",
            is_oneshot: bool = False
    ) -> None:
        """use empty str as wildcard for sender, receiver and message_type"""
        if (sender_id, receiver_id, message_type, replying_to_msg) not in self.__message_callbacks:
            self.__message_callbacks[(sender_id, receiver_id, message_type, replying_to_msg)] = []
        self.__message_callbacks[(sender_id, receiver_id, message_type, replying_to_msg)].append((callback, is_oneshot))

    def remove_message_callback(
            self,
            callback: MsgCallbackType,
            sender_id: str = "",
            receiver_id: str = "",
            message_type: str = "",
            replying_to_msg: str = "",
            is_oneshot: bool = False
    ) -> None:
        """use empty str as wildcard for sender, receiver and message_type"""
        if (sender_id, receiver_id, message_type, replying_to_msg) in self.__message_callbacks:
            self.__message_callbacks[(sender_id, receiver_id, message_type, replying_to_msg)].remove((callback, is_oneshot))
            if len(self.__message_callbacks[(sender_id, receiver_id, message_type, replying_to_msg)]) == 0:
                self.__message_callbacks.pop((sender_id, receiver_id, message_type, replying_to_msg))

    @staticmethod
    def __check_filter(msg_filter: Tuple[str, ...], val: Tuple[str, ...]) -> bool:
        """checks if val fits filter, empty string as wildcard"""
        if len(msg_filter) != len(val):
            raise ValueError("length of filter != length of value")
        for i in range(len(msg_filter)):
            if msg_filter[i] != "":
                if msg_filter[i] != val[i]:
                    return False
        return True

    def process_message(self, message: Dict[str, Any], loop=None):
        replying_to_msg = ""
        sender_id = ""
        receiver_id = ""
        message_type = ""
        try:
            sender_id = message["sender"]
            receiver_ids = message["receivers"]
            message_type = message["messageType"]
            if "replying_to_msg" in message:
                replying_to_msg = message["replying_to_msg"]
            message_signatures = [(sender_id, r_id, message_type, replying_to_msg) for r_id in receiver_ids]
        except KeyError as e:
            self.logger.error(f"Keyerror accessing message for callback filter, message: {message}, error: {e} ")
            return
        if loop is None:
            loop = asyncio.get_running_loop()
        for msg_filter, method_list in self.__message_callbacks.items():
            for signature in message_signatures:
                if self.__check_filter(msg_filter, signature):
                    for method, is_oneshot in method_list:
                        try:
                            loop.create_task(method(message))
                        except:
                            self.logger.error(f"unexpected error trying to create callback task for message {message}")
                        if is_oneshot:
                            self.remove_message_callback(
                                callback=method,
                                sender_id=sender_id,
                                receiver_id=receiver_id,
                                message_type=message_type,
                                is_oneshot=is_oneshot
                            )




import gzip
import json
import gnupg
import pgpy
from abc import ABC
from s3i.exception import raise_error_from_s3ib_msg, S3IBMessageError
from typing import Union, Optional, Dict, List
GZIP_MAGIC_NUMBER = "1f8b"


class TimeInterval:
    def __init__(self, start: int, end: int):
        self.__start = start
        self.__end = end

    @property
    def start(self) -> int:
        return self.__start

    @start.setter
    def start(self, value: int):
        self.__start = value

    @property
    def end(self) -> int:
        return self.__end

    @end.setter
    def end(self, value: int):
        self.__end = value


class Message(ABC):
    def __init__(self, base_msg=None, gzip_msg=None, pgp_msg=None):
        if isinstance(base_msg, dict):
            raise_error_from_s3ib_msg(base_msg, S3IBMessageError)
            self.__base_msg = base_msg

        elif isinstance(base_msg, str):
            base_msg_json = json.loads(base_msg)
            raise_error_from_s3ib_msg(base_msg_json, S3IBMessageError)
            self.__base_msg = base_msg_json

        elif isinstance(base_msg, bytes):
            base_msg = str(base_msg, 'utf-8')
            base_msg_json = json.loads(base_msg)
            raise_error_from_s3ib_msg(base_msg_json, S3IBMessageError)
            self.__base_msg = base_msg_json

        else:
            self.__base_msg = {}

        if isinstance(pgp_msg, (str, bytes)):
            self.__pgp_msg = pgpy.pgp.PGPMessage.from_blob(pgp_msg)
        elif isinstance(pgp_msg, pgpy.pgp.PGPMessage):
            self.__pgp_msg = pgp_msg
        else:
            self.__pgp_msg = pgpy.PGPMessage.new("")

        if isinstance(gzip_msg, bytes):
            if gzip_msg.hex()[0:4] == GZIP_MAGIC_NUMBER:
                self.__gzip_msg = gzip_msg
            else:
                raise ValueError("Invalid gzip header")
        else:
            self.__gzip_msg = b''

    @property
    def base_msg(self):
        return self.__base_msg

    @base_msg.setter
    def base_msg(self, value):
        if isinstance(value, dict):
            raise_error_from_s3ib_msg(value, S3IBMessageError)
            self.__base_msg = value
        else:
            raise TypeError("Only dict typed message is valid")

    @property
    def pgp_msg(self):
        return self.__pgp_msg

    @pgp_msg.setter
    def pgp_msg(self, value):
        if isinstance(value, pgpy.pgp.PGPMessage):
            self.__pgp_msg = value
        else:
            raise TypeError("Only message typed with pgpy.pgp.PGPMessage is valid")

    @property
    def gzip_msg(self):
        return self.__gzip_msg

    @gzip_msg.setter
    def gzip_msg(self, value):
        if isinstance(value, bytes):
            if value.hex()[0:4] == GZIP_MAGIC_NUMBER:
                self.__gzip_msg = value
            else:
                raise ValueError("Invalid gzip header")
        else:
            raise TypeError("Only bytes typed message is valid")

    def convert_base_to_pgp(self):
        self.pgp_msg = pgpy.PGPMessage.new(json.dumps(self.__base_msg))
        return self.pgp_msg

    def convert_pgp_to_base(self):
        if self.pgp_msg.is_encrypted:
            raise ValueError("PGP Message is still encrypted")
        else:
            self.base_msg = json.loads(self.__pgp_msg.message)
            return self.base_msg

    def convert_gzip_to_pgp(self):
        self.pgp_msg = pgpy.PGPMessage.new(self.__gzip_msg)
        return self.pgp_msg

    def convert_pgp_to_gzip(self):
        if self.pgp_msg.is_encrypted:
            raise ValueError("PGP Message is still encrypted")
        else:
            if not isinstance(self.pgp_msg.message, bytearray):
                raise ValueError("PGP Message can not be converted to GZIP data")
            self.gzip_msg = bytes(self.pgp_msg.message)
            return self.gzip_msg

    def sign(self, sec_key, msg_pgp):
        if not isinstance(sec_key, type(pgpy.PGPKey())):
            raise TypeError("Invalid type of secret key")
        if not isinstance(msg_pgp, pgpy.pgp.PGPMessage):
            raise TypeError("Only PGP Message is valid")
        msg_pgp |= sec_key.sign(msg_pgp)
        self.pgp_msg = msg_pgp
        return self.pgp_msg

    @staticmethod
    def verify(pub_key, msg_pgp):
        if not isinstance(pub_key, type(pgpy.PGPKey())):
            raise TypeError("Invalid type of public key")
        if not isinstance(msg_pgp, pgpy.pgp.PGPMessage):
            raise TypeError("Only PGP Message is valid")
        try:
            pub_key.verify(msg_pgp)
            return True
        except pgpy.errors.PGPError:
            return False

    def encrypt(self, pub_key, msg_pgp):
        if not isinstance(pub_key, type(pgpy.PGPKey())):
            raise TypeError("Invalid type of public key")
        if not isinstance(msg_pgp, pgpy.pgp.PGPMessage):
            raise TypeError("Only PGP Message is valid")
        cipher = pgpy.constants.SymmetricKeyAlgorithm.AES256
        session_key = cipher.gen_key()
        self.pgp_msg = pub_key.encrypt(msg_pgp, cipher=cipher, sessionkey=session_key)  # TODO: double print of logs
        return self.pgp_msg

    def decrypt(self, sec_key, msg_pgp):
        if not isinstance(sec_key, type(pgpy.PGPKey())):
            raise TypeError("Invalid type of public key")
        if not isinstance(msg_pgp, pgpy.pgp.PGPMessage):
            raise TypeError("Only PGP Message is valid")
        self.pgp_msg = sec_key.decrypt(msg_pgp)
        return self.pgp_msg

    def compress(self, msg_json, level):
        if not isinstance(level, int):
            raise TypeError("Unexpected type for the input level")
        self.gzip_msg = gzip.compress(bytes(json.dumps(msg_json), 'utf-8'), compresslevel=level)
        return self.gzip_msg

    def decompress(self, msg_gzip):
        if not isinstance(msg_gzip, bytes):
            raise TypeError("Unexpected type")
        if msg_gzip.hex()[0:4] != GZIP_MAGIC_NUMBER:
            raise TypeError("Unexpected gzip header")
        self.base_msg = json.loads(gzip.decompress(msg_gzip))
        return self.base_msg


class UserMessage(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillUserMessage(self,
                        sender,
                        receivers,
                        subject,
                        text,
                        message_id,
                        reply_to_endpoint=None,
                        replying_to_msg=None,
                        attachments=None
                        ):

        _user_message_json = {
            "sender": sender,
            "identifier": message_id,
            "receivers": receivers,
            "messageType": "userMessage",
            "subject": subject,
            "text": text
        }
        if replying_to_msg is not None:
            _user_message_json["replyingToMessage"] = replying_to_msg
        if reply_to_endpoint is not None:
            _user_message_json["replyToEndpoint"] = reply_to_endpoint
        if isinstance(attachments, list):
            _user_message_json["attachments"] = [
                {"filename": "", "data": ""} for att in attachments
            ]
            for i in range(len(attachments)):
                _user_message_json["attachments"][i]["filename"] = attachments[i]["filename"]
                _user_message_json["attachments"][i]["data"] = attachments[i]["data"]

        self.base_msg = _user_message_json


class ServiceRequest(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillServiceRequest(self,
                           sender,
                           receivers,
                           message_id,
                           service_type,
                           parameters=None,
                           reply_to_endpoint=None,
                           replying_to_msg=None,
                           ):
        _service_req_json = {
            "sender": sender,
            "identifier": message_id,
            "receivers": receivers,
            "messageType": "serviceRequest",
            "serviceType": service_type,
        }
        if replying_to_msg is not None:
            _service_req_json["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _service_req_json["replyToEndpoint"] = reply_to_endpoint

        if parameters is not None:
            _service_req_json["parameters"] = parameters

        self.base_msg = _service_req_json


class ServiceReply(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillServiceReply(self,
                         sender,
                         receivers,
                         message_id,
                         service_type,
                         results=None,
                         reply_to_endpoint=None,
                         replying_to_msg=None,
                         ):
        _service_reply_json = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "messageType": "serviceReply",
            "serviceType": service_type
        }

        if replying_to_msg is not None:
            _service_reply_json["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _service_reply_json["replyToEndpoint"] = reply_to_endpoint

        if results is not None:
            _service_reply_json["results"] = results

        self.base_msg = _service_reply_json


class GetValueRequest(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillGetValueRequest(self,
                            sender,
                            receivers,
                            message_id,
                            attribute_path,
                            reply_to_endpoint=None,
                            replying_to_msg=None,
                            time_interval: Union[TimeInterval,Dict[str, int]] = None
                            ):

        _get_value_req = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "attributePath": attribute_path,
            "messageType": "getValueRequest"
        }

        if replying_to_msg is not None:
            _get_value_req["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _get_value_req["replyToEndpoint"] = reply_to_endpoint

        if time_interval is not None:
            if type(time_interval) == dict:
                _get_value_req["timeInterval"] = {
                    "start": int(time_interval["start"]),
                    "end": int(time_interval["end"])
                }
            elif type(time_interval) == TimeInterval:
                _get_value_req["timeInterval"] = {
                    "start": time_interval.start,
                    "end": time_interval.end
                }
            else:
                raise TypeError()

        self.base_msg = _get_value_req


class GetValueReply(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillGetValueReply(self,
                          sender,
                          receivers,
                          message_id,
                          value,
                          reply_to_endpoint=None,
                          replying_to_msg=None,
                          ):
        _get_value_reply = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "value": value,
            "messageType": "getValueReply"
        }
        if replying_to_msg is not None:
            _get_value_reply["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _get_value_reply["replyToEndpoint"] = reply_to_endpoint

        self.base_msg = _get_value_reply


class DeleteAttributeRequest(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillDeleteAttributeRequest(self,
                            sender,
                            receivers,
                            message_id,
                            attribute_path,
                            reply_to_endpoint=None,
                            replying_to_msg=None,
                            time_interval: Union[TimeInterval,Dict[str, int]] = None
                            ):

        _delete_attribute_req = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "attributePath": attribute_path,
            "messageType": "deleteAttributeRequest"
        }

        if replying_to_msg is not None:
            _delete_attribute_req["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _delete_attribute_req["replyToEndpoint"] = reply_to_endpoint

        if time_interval is not None:
            if type(time_interval) == dict:
                _delete_attribute_req["timeInterval"] = {
                    "start": int(time_interval["start"]),
                    "end": int(time_interval["end"])
                }
            elif type(time_interval) == TimeInterval:
                _delete_attribute_req["timeInterval"] = {
                    "start": time_interval.start,
                    "end": time_interval.end
                }
            else:
                raise TypeError()

        self.base_msg = _delete_attribute_req


class DeleteAttributeReply(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillDeleteAttributeReply(self,
                          sender: str,
                          receivers: List[str],
                          message_id: str,
                          ok: bool,
                          reply_to_endpoint: str = None,
                          replying_to_msg: str = None,
                          ):
        _delete_attribute_reply = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "ok": ok,
            "messageType": "deleteAttributeReply"
        }
        if replying_to_msg is not None:
            _delete_attribute_reply["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _delete_attribute_reply["replyToEndpoint"] = reply_to_endpoint

        self.base_msg = _delete_attribute_reply



class SetValueRequest(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillSetValueRequest(self,
                            sender,
                            receivers,
                            message_id,
                            attribute_path,
                            new_value,
                            reply_to_endpoint=None,
                            replying_to_msg=None,
                            ):

        _set_value_req = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "attributePath": attribute_path,
            "newValue": new_value,
            "messageType": "setValueRequest"
        }

        if replying_to_msg is not None:
            _set_value_req["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _set_value_req["replyToEndpoint"] = reply_to_endpoint

        self.base_msg = _set_value_req


class SetValueReply(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillSetValueReply(self,
                          sender,
                          receivers,
                          message_id,
                          ok,
                          reply_to_endpoint=None,
                          replying_to_msg=None,
                          ):
        _set_value_reply = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "ok": ok,
            "messageType": "setValueReply"
        }

        if replying_to_msg is not None:
            _set_value_reply["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _set_value_reply["replyToEndpoint"] = reply_to_endpoint

        self.base_msg = _set_value_reply


class SubscribeCustomEventRequest(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillSubscribeCustomEventRequest(self,
                                        sender,
                                        receivers,
                                        message_id,
                                        filter,
                                        attribute_paths,
                                        reply_to_endpoint=None,
                                        replying_to_msg=None,
                                        ):

        _subscribe_custom_event_req = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "filter": filter,
            "attributePaths": attribute_paths,
            "messageType": "subscribeCustomEventRequest"
        }
        if replying_to_msg is not None:
            _subscribe_custom_event_req["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _subscribe_custom_event_req["replyToEndpoint"] = reply_to_endpoint

        self.base_msg = _subscribe_custom_event_req


class SubscribeCustomEventReply(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillSubscribeCustomEventReply(self,
                                      sender,
                                      receivers,
                                      message_id,
                                      topic,
                                      status,
                                      reply_to_endpoint=None,
                                      replying_to_msg=None,
                                      ):
        _subscribe_custom_event_reply = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "topic": topic,
            "status": status,
            "messageType": "subscribeCustomEventReply"
        }

        if replying_to_msg is not None:
            _subscribe_custom_event_reply["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _subscribe_custom_event_reply["replyToEndpoint"] = reply_to_endpoint

        self.base_msg = _subscribe_custom_event_reply


class UnsubscribeCustomEventRequest(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillUnsubscribeCustomEventRequest(self,
                                          sender,
                                          receivers,
                                          message_id,
                                          topic,
                                          reply_to_endpoint=None,
                                          replying_to_msg=None,
                                          ):

        _unsubscribe_custom_event_req = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "topic": topic,
            "messageType": "unsubscribeCustomEventRequest"
        }
        if replying_to_msg is not None:
            _unsubscribe_custom_event_req["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _unsubscribe_custom_event_req["replyToEndpoint"] = reply_to_endpoint

        self.base_msg = _unsubscribe_custom_event_req


class UnsubscribeCustomEventReply(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillUnsubscribeCustomEventReply(self,
                                        sender,
                                        receivers,
                                        message_id,
                                        topic,
                                        status,
                                        reply_to_endpoint=None,
                                        replying_to_msg=None,
                                        ):
        _unsubscribeCustomEventReply = {
            "sender": sender,
            "receivers": receivers,
            "identifier": message_id,
            "topic": topic,
            "status": status,
            "messageType": "unsubscribeCustomEventReply"
        }

        if replying_to_msg is not None:
            _unsubscribeCustomEventReply["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _unsubscribeCustomEventReply["replyToEndpoint"] = reply_to_endpoint

        self.base_msg = _unsubscribeCustomEventReply


class EventMessage(Message):
    def __init__(self, base_msg=None, pgp_msg=None, gzip_msg=None):
        super().__init__(
            base_msg=base_msg,
            pgp_msg=pgp_msg,
            gzip_msg=gzip_msg
        )

    def fillEventMessage(self,
                         sender,
                         message_id,
                         topic: str,
                         timestamp: int,
                         content: dict,
                         reply_to_endpoint=None,
                         replying_to_msg=None,
                         ):

        _event_message = {
            "sender": sender,
            "identifier": message_id,
            "topic": topic,
            "timestamp": timestamp,
            "content": content,
            "messageType": "eventMessage"
        }

        if replying_to_msg is not None:
            _event_message["replyingToMessage"] = replying_to_msg

        if reply_to_endpoint is not None:
            _event_message["replyToEndpoint"] = reply_to_endpoint

        self.base_msg = _event_message



import requests
import json
from enum import Enum
import time
import threading
from jwt.algorithms import RSAPSSAlgorithm
from jwt import DecodeError, ExpiredSignature, InvalidTokenError
import jwt
from s3i.exception import S3IIdentityProviderError, raise_error_from_response
from s3i.logger import APP_LOGGER
import webbrowser


oauth2_flows = ["password", "client_credentials", "authorization_code_flow"]
oauth_proxy_url = "https://auth.s3i.vswf.dev"

class TokenType(Enum):
    """Enum TokenType covers all available tokens specified in openID connect"""

    # COMPLETE_BUNDLE = 1
    ID_TOKEN = 2
    ACCESS_TOKEN = 3
    REFRESH_TOKEN = 4


class IdentityProvider:
    """ Class IdentityProvider contains functions to communicate with S3I IdentityProvider """

    def __init__(self, grant_type, client_id, client_secret, realm="KWH",
                 identity_provider_url="https://idp.s3i.vswf.dev/", username=None,
                 password=None):
        """
        Constructor

        :param grant_type: grant type to obtain JWT. Here only the types "password" and "client_credentials" are implemented. In subsequent version authorization_code will be implemented 
        :type grant_type: str
        :param client_id: id of registered client in S3I IdentityProvider
        :type client_id: str 
        :param client_secret: credential of registered client in S3I IdentityProvider which must corredspond to client id 
        :type client_secret: str 
        :param realm: registered realm in S3I IdentityProvider 
        :type realm: str
        :param identity_provider_url: url of S3I IdentityProvider 
        :type identity_provider_url: str 
        :param username: username of registered user in S3I IdentityProvider. If grant type is set up as password, this field must be filled 
        :type username: str 
        :param password: password of registered user in S3I IdentityProvider. If grant type is set up as password, this field must be filled 
        :type password: str

        """
        self._grant_type = grant_type
        self._client_id = client_id
        self._client_secret = client_secret
        self._realm = realm
        self._username = username
        self._password = password
        self._identity_provider_url = identity_provider_url
        self._identity_provider_get_token = self.identity_provider_url + \
                                            "auth/realms/" + self.realm + "/protocol/openid-connect/token"
        self._identity_provider_header = {
            "Content-Type": "application/x-www-form-urlencoded"}
        self._token_bundle = None
        self._token_inspector_run = False
        self._last_login = 0
        self._identity_provider_get_pub_key = self.identity_provider_url + \
                                              "auth/realms/" + self.realm + "/protocol/openid-connect/certs"

    @property
    def identity_provider_url(self):
        """Url of S3I IdentityProvider
        """
        return self._identity_provider_url

    @property
    def identity_provider_get_token(self):
        """Url to obtain a token from the S3I IdentityProvider
        """
        return self._identity_provider_get_token

    @property
    def identity_provider_header(self):
        """Header which is sent to the S3I IdentityProvider via HTTP
        """
        return self._identity_provider_header

    @property
    def grand_type(self):
        """OAuth grant type which is used
        """
        return self._grant_type

    @property
    def client_id(self):
        """
        """
        return self._client_id

    @property
    def client_secret(self):
        """
        """
        return self._client_secret

    @property
    def realm(self):
        """
        """
        return self._realm

    @property
    def username(self):
        """
        """
        return self._username

    @property
    def password(self):
        """
        """
        return self._password

    def stop_run_forever(self):
        """ Stops the run forever loop when the next iteration happens """
        self._token_inspector_run = False

    def get_token(self, token_type, request_new=False, scope="openid"):
        """ 
        Returns a token from the S3I IdentityProvider, request a new one if there is no valid token available
        Works only if the IdentityProvider is NOT in the run_forever loop

        :param token_type: type of token, see enum TokenType
        :type token_type: enum
        :param request_new: request a new token in any case
        :type request_new: bool
        :param scope: client scope
        :type scope: str
        :return: token
        :rtype: str
        """
        if self._token_inspector_run:
            return None
        if request_new:
            self._authenticate(scope=scope)
        else:
            if self._token_bundle is None:
                self._authenticate(scope=scope)
            if self._time_until_token_valid() <= 0:
                self._authenticate(scope=scope)
        # all tokens are valid
        if token_type == TokenType.ACCESS_TOKEN:
            return self._token_bundle["access_token"]
        elif token_type == TokenType.ID_TOKEN:
            return self._token_bundle["id_token"]
        elif token_type == TokenType.REFRESH_TOKEN:
            return self._token_bundle["refresh_token"]
        return None

    def get_certs(self):
        """
        Return a certificate of S³I Identity Provider

        """
        certs = requests.get(url=self._identity_provider_get_pub_key,
                             headers={"Authorization": "Bearer {}".format(
                                 self.get_token(token_type=TokenType.ACCESS_TOKEN))})
        return certs.json()

    def verify_token(self, token, aud="rabbitmq"):
        """
        This function verifies if the given jwt token is issued by the S³I Identity Provider
 
        :param token: jwt token
        :type token: str
        :param aud: audience, by default rabbitmq.
        :type aud: str
        :return: True if token is issued by S³I, else False
        :rtype: bool
        """
        keys = self.get_certs().get("keys")
        """search public key"""
        for key in keys:
            if key["alg"] == "RS256" and key["kty"] == "RSA" and key["use"] == "sig":
                public_key = RSAPSSAlgorithm.from_jwk(json.dumps(key))
                break
        try:
            """decode the access token with public key that was searched above"""
            jwt.decode(jwt=token, verify=True, key=public_key, algorithms=['RS256'], audience=aud)
        except (DecodeError, ExpiredSignature, InvalidTokenError):
            return False
        else:
            return True

    def run_forever(self, token_type, on_new_token, sleep_interval=5, scope=""):
        """ 
        Requests tokens from the S³I IdentityProvider and refreshs them if they reach their timeout

        :param token_type: type of token, see enum TokenType (REFRESH_TOKEN is no valid param)
        :type token_type: enum
        :param on_new_token: callback if a new token is available
        :type on_new_token: callback
        :param scope: client scope
        :type scope: str
        """
        if token_type == TokenType.REFRESH_TOKEN:
            return None
        on_new_token(self.get_token(token_type, request_new=True, scope=scope))
        self._token_inspector_run = True
        threading._start_new_thread(
            self._run_forever_loop, (token_type, on_new_token, sleep_interval, scope))

    def _run_forever_loop(self, token_type, on_new_token, slep_interval, scope=""):
        while True and self._token_inspector_run:
            while True:
                if not self._token_inspector_run:
                    return None  # exit
                if self._time_until_token_valid() > slep_interval * 2:
                    time.sleep(slep_interval)
                else:
                    break
            # its time to refresh the tokens
            APP_LOGGER.info("request for new token")
            APP_LOGGER.info("last login: {}, now: {}, expires in: {}".format(self._last_login,
                                                                             time.time(),
                                                                             self._token_bundle["expires_in"]))
            if self._token_bundle["expires_in"] < 60:
                ### keycloak returns an unexpected token expire time
                self._authenticate(scope=scope)
            else:
                self._refresh_token(self._token_bundle["refresh_token"], scope)
            if token_type == TokenType.ACCESS_TOKEN:
                on_new_token(self._token_bundle["access_token"])
            elif token_type == TokenType.ID_TOKEN:
                on_new_token(self._token_bundle["id_token"])

    def _authenticate(self, scope):
        """ Request the token-bundle from the S³I IdentityProvider 
        
            :param scope: client scope
            :type scope: str 
        """

        self._last_login = time.time()
        if self.grand_type not in oauth2_flows:
            return {}
        elif self.grand_type == "authorization_code_flow":
            init_url = "{0}/initialize/{1}/{2}".format(oauth_proxy_url, self.client_id, self.client_secret)
            init_resp = requests.get(init_url)
            init_resp_json = init_resp.json()
            webbrowser.open_new_tab("{}/{}".format(oauth_proxy_url, init_resp_json["redirect_url"]))
            proxy_user_id = init_resp_json["proxy_user_identifier"]
            proxy_user_secret = init_resp_json["proxy_secret"]
            pickup_url = "{}/pickup/{}/{}".format(oauth_proxy_url, proxy_user_id, proxy_user_secret)

            max_retries = 100
            pickup_resp = requests.get(pickup_url)
            for i in range(max_retries):
                if not pickup_resp.text:
                    pickup_resp = requests.get(pickup_url)
                    time.sleep(0.5)
                else:
                    break
            pickup_resp_json = pickup_resp.json()
            decoded_jwt = jwt.decode(pickup_resp_json["access_token"], verify=False)
            pickup_resp_json["expires_in"] = decoded_jwt["exp"] - decoded_jwt["iat"]
            self._token_bundle = pickup_resp_json

        else:
            body = dict()
            body["grant_type"] = self._grant_type
            body["client_id"] = self._client_id
            body["client_secret"] = self._client_secret
            if self.grand_type == "password":
                body["username"] = self._username
                body["password"] = self._password
            if scope:
                body["scope"] = scope
            response = requests.post(url=self._identity_provider_get_token,
                                     data=body, headers=self._identity_provider_header)
            response_json = raise_error_from_response(response, S3IIdentityProviderError, 200)
            self._token_bundle = response_json

    def _refresh_token(self, token, scope=""):
        """ Refreshes the given token

        :param token: type of token, see enum TokenType
        :type token: str
        """
        self._last_login = time.time()
        body = dict()
        body["grant_type"] = "refresh_token"
        body["client_id"] = self._client_id
        body["client_secret"] = self._client_secret
        body["refresh_token"] = token
        if scope:
            body["scope"] = scope
        response = requests.post(url=self._identity_provider_get_token,
                                 data=body, headers=self._identity_provider_header)
        response_json = raise_error_from_response(response, S3IIdentityProviderError, 200)
        self._token_bundle = response_json

    def _time_until_token_valid(self):
        """ Returns the time until the token expires """
        time_token_valid = self._token_bundle["expires_in"]
        time_since_login = time.time() - self._last_login
        return time_token_valid - time_since_login



import urllib

import requests

from s3i.exception import S3IDittoError, raise_error_from_response


def query_all(fn):
    """Adds pagination to a request.

    Expects a function, that returns a rest get request without the keyword "cursor".
    Returns a list of json objects, that meet the query.

    :param fn: Function returning a rest get request
    :type fn: function
    """
    def inner(*args, **kwargs):
        url, headers = fn(*args, **kwargs)
        response = requests.get(url, headers=headers)
        response_json = raise_error_from_response(response, S3IDittoError, 200)
        results = response_json["items"]
        cursor = response_json.get("cursor", None)
        while cursor:
            response = requests.get(url,
                                    headers=headers,
                                    params={"option": f"cursor({cursor})"})
            response_json = raise_error_from_response(response, S3IDittoError, 200)
            items = response_json["items"]
            results += items
            cursor = response_json.get("cursor", None)
        return results

    return inner

def http_error_from_request_response(response):
    """Returns an urllib.error.HTTPError created from a requests.Response
    object.

    :type response: requests.Response
    :rtype: urllib.error.HTTPError

    """
    error = urllib.error.HTTPError(
        code=response.status_code,
        msg=response.text,
        hdrs=response.headers,
        fp=None,
        url=response.url,
    )
    return error



import json
from urllib.parse import urljoin

import requests

from s3i.exception import S3IDittoError, raise_error_from_response
from s3i.tools import query_all


class DittoManager:
    """Class Directory contains functions to query things in S3I Directory """

    def __init__(self, ditto_url, token):
        """
        Constructor

        :param ditto_url: url of S3I Directory
        :type ditto_url: str
        :param token: id token obtained from S3I IdP
        :type token: str
        """

        self.__ditto_url = ditto_url
        self.__token = token
        """headers to bearer authentication at S3I Directory"""
        self.__ditto_headers = {
            'Content-Type': 'application/json',
            'Authorization': 'Bearer ' + self.__token
        }

    @property
    def ditto_url(self):
        """Url of S3I Directory
        """
        return self.__ditto_url

    @ditto_url.setter
    def ditto_url(self, value):
        self.__ditto_url = value

    @property
    def token(self):
        """Access Token from user which will be sent within HTTP request
        """
        return self.__token

    @token.setter
    def token(self, value):
        self.__token = value

    @property
    def ditto_headers(self):
        """HTTP Header which is sent within HTTP request
        """
        return self.__ditto_headers

    @ditto_headers.setter
    def ditto_headers(self, value):
        self.__ditto_headers = value

    def pass_refreshed_token(self, token):
        """
        If token expires, it is necessary to pass a refreshed token using this method.
        """
        self.__token = token
        self.__ditto_headers = {
            'Content-Type': 'application/json',
            'Authorization': 'Bearer ' + self.__token
        }

    def updateThingIDBased(self, thingID, payload):
        """
        Update a thing's entry using its thingId

        :param thingID: id of the thing
        :type thingID: str
        :param payload: thing description to update in Dir
        :type payload: dict
        :return: requests.Response() see: ResponseObject_
        """
        response = requests.put(url=self.ditto_url + "things/{}".format(thingID),
                                data=json.dumps(payload), headers=self.ditto_headers)
        raise_error_from_response(response, S3IDittoError, 204)
        return response

    def queryThingIDBased(self, thingID):
        """Query a thing's entry using its ID

        :param thingID: id of the thing
        :type thingID: str
        :return: response of http requests
        :rtype: requests.Response() see: ResponseObject_
        """
        response = requests.get(
            url=self.ditto_url + "things/" + thingID, headers=self.ditto_headers)
        response_json = raise_error_from_response(response, S3IDittoError, 200)
        return response_json

    @query_all
    def queryAttributeBased(self, key, value, filter="eq"):
        """Query a thing's entry based on one of its attribute's value

        :param key: fully qualified name of the relevant attribute, e.g., "thingStructure/links/target/values/value"
        :type key: str
        :param value: searched value of this attribute
        :type value: str
        :return: list of things, which meet the filter requirements
        :rtype: list
        """
        url = urljoin(self.ditto_url, f"search/things?filter={filter}(attributes/{key},\"{value}\")")
        return url, self.ditto_headers

    @query_all
    def searchAll(self, queryParameter=None):
        """List all thing entries in S3I Directory which are reachable for user who is represented in the token.

        :return: list of all thing entries which are reachable for user who is represented in the token
        :rtype: list
        """
        url = urljoin(self.ditto_url, "search/things/")
        if queryParameter is not None:
            url = urljoin(url, f",filter={queryParameter}")

        return url, self.ditto_headers

    def updatePolicyIDBased(self, policyID, payload):
        """
        Update a policy's entry using its ID

        :param policyID: id of the thing
        :type policyID: str
        :param payload: thing description to update in Dir
        :type payload: dict
        :return: requests.Response() see: ResponseObject_
        """
        response = requests.put(url=self.ditto_url + "policies/" + policyID,
                                data=json.dumps(payload), headers=self.ditto_headers)
        raise_error_from_response(response, S3IDittoError, 204)
        return response

    def queryPolicyIDBased(self, policyID):
        """Query a policy's entry using its ID

        :param policyID: id of the thing
        :type policyID: str
        :return: response of http requests
        :rtype: requests.Response() see: ResponseObject_
        """
        response = requests.get(
            url=self.ditto_url + "policies/" + policyID, headers=self.ditto_headers)
        response_json = raise_error_from_response(response, S3IDittoError, 200)
        return response_json



import asyncio
import functools
import ssl
import requests
import urllib
import json

import pika
import pika.exceptions
from pika.adapters.asyncio_connection import AsyncioConnection
from s3i.exception import raise_error_from_response, raise_error_from_s3ib_amqp, S3IBrokerRESTError, S3IBrokerAMQPError
from s3i.logger import APP_LOGGER
from s3i.callback_manager import CallbackManager
from s3i.config import Config

CONTENT_TYPE = "application/json"
HOST = "broker.s3i.vswf.dev"
VIRTUAL_HOST = "s3i"
DIRECT_EXCHANGE = "demo.direct"
EVENT_EXCHANGE = "eventExchange"


class BrokerREST:
    """
    Class Broker REST contains functions to connect to S3I Broker via HTTP REST API, and send and receive messages

    """

    def __init__(self, token, url="https://broker.s3i.vswf.dev/"):
        """
        Constructor

        :param token: Access Token issued from S³I IdentityProvider
        :type token: str
        :param url: url of S³I Broker API
        :type url: str

        """
        self._token = token
        self._url = url
        self.headers = {'Content-Type': 'application/json',
                        'Authorization': 'Bearer ' + self.token}
        self.headers_encrypted = {'Content-Type': 'application/pgp-encrypted',
                                  'Authorization': 'Bearer ' + self.token}

    @property
    def token(self):
        """Returns the JWT currently in use.

        :returns: JWT-Token
        :rtype: str

        """
        return self._token

    @token.setter
    def token(self, new_token):
        """Sets the value of the object's token property to new_token.

        :param new_token: JWT
        :type new_token: str
        """
        self._token = new_token
        self.headers["Authorization"] = "Bearer " + new_token
        self.headers_encrypted["Authorization"] = "Bearer " + new_token

    def send(self, endpoints, msg, encrypted=False):
        """
        Send a S³I-B message via S³I Broker API
        :param endpoints: endpoints of the receivers
        :type endpoints: list of str
        :param msg: message to be sent
        :type msg: str
        :param encrypted: if true, message will be sent encrypted, otherwise not.
        :type encrypted: bool

        """
        if encrypted:
            headers = self.headers_encrypted
        else:
            headers = self.headers
        for endpoint in endpoints:
            response = requests.post(url="{}/{}".format(self._url, endpoint), headers=headers, data=msg)
            raise_error_from_response(response, S3IBrokerRESTError, 201)
        return True

    def receive_once(self, queue):
        """
        Receive one S3I-B message and do not wait for more messages.

        :param queue: queue which starts a listener in order to receive a single message
        :type queue: str
        :return: received S3I-B message
        :rtype: dict
        """

        queue_encoded = urllib.parse.quote(queue, safe="")
        url = "{}{}".format(self._url, queue_encoded)
        response = requests.get(url=url, headers=self.headers)
        response_json = raise_error_from_response(response, S3IBrokerRESTError, 200)
        return response_json


class BrokerAMQP:
    _ON_CONNECTION_OPEN = "_on_connection_open"
    _ON_CONNECTION_CLOSED = "_on_connection_closed"
    _ON_CHANNEL_OPEN = "_on_channel_open"
    _ON_CHANNEL_CLOSED = "_on_channel_closed"

    def __init__(self, token, endpoint, callback, loop=asyncio.get_event_loop(), listenToEvents=False):
        self.__token = token
        self.__endpoint = endpoint
        self.__event_endpoint = None
        self.__loop = loop
        self.__callback = callback
        self.__credentials = None
        self.__connection_parameters = None
        self.__connection = None
        self.__channel = None
        self.__consumer_tag = None
        self.__event_consumer_tag = None
        self.__listen_to_events = listenToEvents

        self.__is_consuming = False
        self.__schedule_messages = []
        self.__callbacks = CallbackManager()

    @property
    def token(self):
        return self.__token

    @token.setter
    def token(self, value):
        self.__token = value

    @property
    def connection(self):
        return self.__connection

    @property
    def channel(self):
        return self.__channel

    def connect(self):
        self.__credentials = pika.PlainCredentials(
            username=" ",
            password=self.__token,
            erase_on_connect=True
        )
        self.__connection_parameters = pika.ConnectionParameters(
            host=HOST,
            virtual_host=VIRTUAL_HOST,
            credentials=self.__credentials,
            heartbeat=10,
            port=5671,
            ssl_options=pika.SSLOptions(ssl.SSLContext())
        )
        self.__connection = AsyncioConnection(
            parameters=self.__connection_parameters,
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed,
            custom_ioloop=self.__loop
        )
        if self.__listen_to_events:
            self.create_event_queue()

    def on_connection_open(self, _unused_connection):
        APP_LOGGER.info("[S3I]: Connection to Broker built")

        self.__channel = _unused_connection.channel(
            on_open_callback=self.on_channel_open
        )
        self.__callbacks.process(
            self._ON_CONNECTION_OPEN,
            self.__loop
        )

    @staticmethod
    def on_connection_open_error(_unused_connection, err):
        APP_LOGGER.error("[S3I]: Connection to broker failed: {}".format(err))

    def on_connection_closed(self, _unused_connection, reason):
        APP_LOGGER.info("[S3I]: Connection to Broker closed: {}".format(reason))
        if self.__is_consuming:
            self.__until_all_closed_and_reconnect()

    def on_channel_open(self, _unused_channel):
        APP_LOGGER.info("[S3I]: Channel open and start consuming messages")
        _unused_channel.add_on_close_callback(self.on_channel_closed)
        _unused_channel.basic_qos(
            prefetch_count=1
        )
        if self.__callback is not None:
            self.start_consuming()
        self.__callbacks.process(
            self._ON_CHANNEL_OPEN,
            self.__loop
        )

    def add_on_channel_open_callback(self, callback, one_shot, *args, **kwargs):
        self.__callbacks.add(
            self._ON_CHANNEL_OPEN,
            callback,
            one_shot,
            False,
            *args,
            **kwargs
        )

    def add_on_connection_open_callback(self, callback, one_shot, *args, **kwargs):
        self.__callbacks.add(
            self._ON_CONNECTION_OPEN,
            callback,
            one_shot,
            False,
            *args,
            **kwargs
        )

    def start_consuming(self):
        self.__consumer_tag = self.__channel.basic_consume(
            auto_ack=True,
            exclusive=True,
            queue=self.__endpoint,
            on_message_callback=self.__callback
        )
        if self.__event_endpoint is not None:
            self.__event_consumer_tag = self.__channel.basic_consume(
                auto_ack=True,
                exclusive=True,
                queue=self.__event_endpoint,
                on_message_callback=self.__callback)

        self.__is_consuming = True

    def stop_consuming(self):
        if self.__event_consumer_tag is not None:
            self.__channel.basic_cancel(self.__event_consumer_tag)

        cb = functools.partial(
            self.on_consumer_cancel_ok, userdata=self.__consumer_tag
        )
        self.__channel.basic_cancel(self.__consumer_tag, cb)
        self.__is_consuming = False

    def on_channel_closed(self, channel, reason):
        APP_LOGGER.info("[S3I]: Channel is closed: {}".format(reason))
        if not self.__connection.is_closed:
            self.__connection.close()

    def on_consumer_cancel_ok(self, _unused_frame, userdata):
        if not self.__is_consuming:
            self.__channel.close()

    def reconnect_token_expired(self, token):
        self.__token = token
        """
        Stop comsuming and invoke the stop function for channel and connection 
        """
        if self.__is_consuming:
            self.stop_consuming()
        """
        Check if the channel and connection are closed 
        """
        self.__until_all_closed_and_reconnect()

    def __until_all_closed_and_reconnect(self):
        if not self.__channel.is_closed or not self.__connection.is_closed:
            self.__loop.call_later(
                0.1,
                self.__until_all_closed_and_reconnect
            )
        else:
            APP_LOGGER.info("[S3I]: Reconnect to Broker")
            self.connect()

    def send(self, endpoints, msg):
        if isinstance(msg, dict):
            msg = json.dumps(msg)
        if self.__channel.is_open:
            try:
                for endpoint in endpoints:
                    raise_error_from_s3ib_amqp(
                        self.__channel.basic_publish,
                        S3IBrokerAMQPError,
                        DIRECT_EXCHANGE,
                        endpoint,
                        msg,
                        pika.BasicProperties(
                            content_type="application/json",
                            delivery_mode=2
                        ))
                    APP_LOGGER.info("[S3I]: Sending message successes")

            except S3IBrokerAMQPError as err:
                APP_LOGGER.error("[S3I]: Sending message failed: {}".format(err))
                return False
            else:
                return True

    def publish_event(self, msg, topic):
        if self.__channel.is_open:
            try:
                raise_error_from_s3ib_amqp(
                    self.__channel.basic_publish,
                    S3IBrokerAMQPError,
                    EVENT_EXCHANGE,
                    topic,
                    msg,
                    pika.BasicProperties(
                        content_type="application/json",
                    ))
            except S3IBrokerAMQPError as err:
                APP_LOGGER.error("[S3I]: Sending event failed: {}".format(err))
                return False
            else:
                return True

    def create_event_queue(self):
        conf = Config(self.__token)
        identifier = self.__endpoint.replace("s3ibs://", '')
        identifier = identifier.replace("s3ib://", '')
        response = conf.create_broker_event_queue(thing_id=identifier, topic=[])
        self.__event_endpoint = response.json()['queue_name']
        return self.__event_endpoint

    def subscribe_topic(self, topic):
        if self.__channel.is_open and self.__event_endpoint is not None:
            try:
                raise_error_from_s3ib_amqp(
                    self.__channel.queue_bind,
                    S3IBrokerAMQPError,
                    exchange=EVENT_EXCHANGE,
                    queue=self.__event_endpoint,
                    routing_key=topic
                )
            except S3IBrokerAMQPError as err:
                APP_LOGGER.error("[S3I]: Topic subscription failed: {}".format(err))
                return False
            else:
                return True
        else:
            APP_LOGGER.error("[S3I]: No event endpoint configured yet")
            return False

    def unsubscribe_topic(self, topic):
        if self.__channel.is_open and self.__event_endpoint is not None:
            try:
                raise_error_from_s3ib_amqp(
                    self.__channel.queue_unbind,
                    S3IBrokerAMQPError,
                    exchange=EVENT_EXCHANGE,
                    queue=self.__event_endpoint,
                    routing_key=topic
                )
            except S3IBrokerAMQPError as err:
                APP_LOGGER.error("[S3I]: Topic unsubscription failed: {}".format(err))
                return False
            else:
                return True
        else:
            APP_LOGGER.error("[S3I]: No event endpoint configured yet")
            return False



import requests
import json
from s3i.exception import S3IError, raise_error_from_response


class Config:
    """class Config contains function to create and delete persons and things"""

    def __init__(self, token, server_url="https://config.s3i.vswf.dev"):
        """
        Constructor

        :param server_url: server url of S3I Config's REST API
        :type server_url: str
        :param token: access token of requester
        :type token: str
        """
        self.__server_url = server_url
        self.__token = token
        """
        Headers for HTTP Requester against S3I Config REST API
        """
        self.__headers = {"Content-Type": "application/json",
                          "Authorization": "Bearer {}".format(self.token)}

    @property
    def server_url(self):
        return self.__server_url

    @server_url.setter
    def server_url(self, value):
        self.__server_url = value

    @property
    def token(self):
        return self.__token

    @token.setter
    def token(self, value):
        self.__token = value

    @property
    def headers(self):
        return self.__headers

    @headers.setter
    def headers(self, value):
        self.__headers = value

    def create_person(self, username, password):
        """

        create person identity and corresponding thing (including policy) in S3I IdentityProvider and Directory

        :param username: selected username for new person
        :type username: str
        :param password: selected password for new person
        :type password: str
        :return: HTTP Response, 201 if OK
        
        """
        
        url = self.server_url + "/persons/"
        body = dict()
        body["password"] = password
        body["username"] = username
        response = requests.post(url=url, headers=self.headers, data=json.dumps(body))
        raise_error_from_response(response, S3IError, 201)
        return response

    def delete_person(self, username):
        """

        delete person identity and all its owned things from S3I IdentityProvider, Directory and Broker

        :param username: username of person
        :type username: str
        :return: HTTP Response, 204 if OK
        
        """
        
        url = self.server_url + "/persons/{}".format(username)
        response = requests.delete(url=url, headers=self.headers)
        raise_error_from_response(response, S3IError, 204)
        return response

    def create_thing(self, body={}):
        """

        create new thing in S3I, including identity, thing entry and policy in Directory

        :param body: thing configuration for S3I IdentityProvider
        :type body: dict
        :return: HTTP Response, 201 if OK
        
        """
        
        url = self.server_url + "/things/"
        if not isinstance(body, dict):
            body = {}
        response = requests.post(url=url, headers=self.headers, data=json.dumps(body))
        raise_error_from_response(response, S3IError, 201)
        return response

    def delete_thing(self, thing_id):
        """

        delete identity, Directory entry and Broker configuration of a thing from S3I IdP, Dir and Broker

        :param thing_id: id of thing
        :type thing_id: str
        :return: HTTP Response, 204 if OK
        
        """
        
        url = self.server_url + "/things/{}".format(thing_id)
        response = requests.delete(url=url, headers=self.headers)
        raise_error_from_response(response, S3IError, 204)
        return response

    def create_cloud_copy(self, thing_id):
        """
        For an existing person identity, a cloud copy can be created in an repository

        :param thing_id: id of thing
        :type thing_id: str
        :return: HTTP Response, 204 if OK
        
        """
        
        url = self.server_url + "/things/{}/repository".format(thing_id)
        response = requests.post(url=url, headers=self.headers)
        raise_error_from_response(response, S3IError, 201)
        return response

    def delete_cloud_copy(self, thing_id):
        """
        For an existing thing identity, a cloud copy can be deleted from S3I Repository

        :param thing_id: id of thing
        :type thing_id: str
        :return: HTTP Response, 204 if OK
        
        """
        
        url = self.server_url + "/things/{}/repository".format(thing_id)
        response = requests.delete(url=url, headers=self.headers)
        raise_error_from_response(response, S3IError, 204)
        return response

    def create_broker_queue(self, thing_id, encrypted=True):
        """ "
        Creates a message queue for the thing specified by thing_id.

        :param thing_id: id of thing
        :type thing_id: str
        :param encrypted: specifies whether endpoint is encrypted or not
        :type encrypted: boolean
        :return: HTTP Response, 204 if OK
        
        """

        url = f"{self.server_url}/things/{thing_id}/broker"
        body = {"encrypted": encrypted}
        response = requests.post(url=url, headers=self.headers, data=json.dumps(body))
        raise_error_from_response(response, S3IError, 201)
        return response

    def delete_broker_queue(self, thing_id):
        """ "
        Deletes the message queue for the thing specified by thing_id.

        :param thing_id: id of thing
        :type thing_id: str
        :return: HTTP Response, 201 if OK
        
        """
        
        url = f"{self.server_url}/things/{thing_id}/broker"
        response = requests.delete(url=url, headers=self.headers)
        raise_error_from_response(response, S3IError, 204)
        return response

    def create_broker_event_queue(self, thing_id, topic, queue_length=None):
        """
        Creates a event message queue for the thing specified by thing_id

        :param thing_id: thing id
        :param topic: event topic
        :type topic: list or str
        :param queue_length: the length of event queue
        :type queue_length: int
        """
        url = f"{self.server_url}/things/{thing_id}/broker/event"
        body = {"topic": topic}
        if queue_length is not None:
            body["queue_length"] = queue_length
        response = requests.post(url=url, headers=self.headers, data=json.dumps(body))
        raise_error_from_response(response, S3IError, 201)
        return response

    def delete_broker_event_queue(self, thing_id):
        """
        Deletes a event message queue

        :param thing_id: thing id
        """
        url = f"{self.server_url}/things/{thing_id}/broker/event"
        response = requests.delete(url=url, headers=self.headers)
        raise_error_from_response(response, S3IError, 204)
        return response




from s3i.identity_provider import IdentityProvider
from s3i.identity_provider import TokenType
from s3i.ditto_manager import DittoManager
from s3i.directory import Directory
from s3i.repository import Repository
from s3i.broker import BrokerREST, BrokerAMQP
from s3i.config import Config
from s3i.broker_message import UserMessage, ServiceReply, ServiceRequest, GetValueReply, GetValueRequest, \
    SetValueRequest, SetValueReply, \
    EventMessage, SubscribeCustomEventRequest, SubscribeCustomEventReply, UnsubscribeCustomEventRequest, UnsubscribeCustomEventReply
from s3i.key_pgp import Key
from s3i.event_system import EventManager, RQL, CustomEvent, NamedEvent
from s3i.exception import S3IError, S3IDittoError, S3IDirectoryError, S3IRepositoryError, S3IIdentityProviderError, \
    S3IBMessageError, S3IBrokerAMQPError, S3IBrokerRESTError, raise_error_from_response, raise_error_from_s3ib_amqp, \
    raise_error_from_s3ib_msg
from s3i.logger import APP_LOGGER, setup_logger
from s3i.callback_manager import CallbackManager



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['s3i']

package_data = \
{'': ['*']}

install_requires = \
['alabaster>=0.7.12',
 'astroid>=2.3.3',
 'asyncio>=3.4.3,<4.0.0',
 'certifi>=2019.11.28',
 'cffi>=1.13.2',
 'chardet>=1.13.2',
 'colorama>=0.4.3',
 'cryptography==3.1',
 'idna>=2.8',
 'jinja2<3.1',
 'jsonschema>=3.2.0',
 'lazy-object-proxy>=1.4.3',
 'markupsafe>=1.1.1',
 'mccabe>=0.6.1',
 'pgpy>=0.5.4,<0.6.0',
 'pika>=1.2.0',
 'pockets>=0.9.1',
 'pycparser>=2.19',
 'pygments>=2.5.2',
 'pyjwt==1.6.1',
 'pynput>=1.7.6,<2.0.0',
 'pyparsing>=2.4.5',
 'pyrql==0.7.0',
 'python-benedict>=0.21.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'python-gnupg>=0.5.0,<0.6.0',
 'requests>=2.26.0',
 'schema>=0.7.1',
 'six>=1.13.0',
 'snowballstemmer>=2.0.0',
 'typed-ast>=1.4.0',
 'urllib3>=1.25.7',
 'uuid>=1.30,<2.0',
 'websocket-client>=1.4.1,<2.0.0',
 'wrapt>=1.11.2']

setup_kwargs = {
    'name': 's3i',
    'version': '0.6.5',
    'description': 'S3I Basic functions',
    'long_description': '# Documentation\nThis package contains everything you need to get started with the S3I. The documentation of the S3I package and its demos can be found [here](https://kwh40.pages.rwth-aachen.de/s3i/).\n\n# Installation\ns3i is available on PyPI:\n\n```\npython -m pip install s3i\n```\n\n# Development\nThe s3i source code is hosted on [gitlab](https://git.rwth-aachen.de/kwh40/s3i). \n\n## Contributing\nFeel free to contribute to this project. For more details on submitting changes see [CONTRIBUTING.md](./CONTRIBUTING.md)',
    'author': 'Kompetenzzentrum Wald und Holz 4.0',
    'author_email': 's3i@kwh40.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://kwh40.pages.rwth-aachen.de/s3i/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from s3i.ditto_manager import DittoManager


class Repository(DittoManager):
    """Class Directory contains functions to query things in S3I Repository """

    def __init__(self, s3i_repo_url, token):
        """
        Constructor

        :param s3i_repo_url: url of S3I Repository
        :type s3i_repo_url: str
        :param token: id token obtained from S3I IdP
        :type token: str
        """
        super().__init__(s3i_repo_url, token)
